package org.objectweb.fractal.bf.hdl.graph;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import org.objectweb.fractal.bf.hdl.reader.ShortestPathExecutor;

import jdsl.graph.api.Graph;
import jdsl.graph.api.Vertex;
import jdsl.graph.api.VertexIterator;

/**
 * Can travel on a graph looking for different things
 * 
 *
 */
public class GraphTraveller
{
  
  Logger log = Logger.getLogger(GraphTraveller.class.getCanonicalName());
  
  private static GraphTraveller INSTANCE;
  
  /**
   * local cache, to avoid recomputation over already visited graphs (maybe
   * premature optimization). 
   */
  private HashMap<Graph, Set<PairOfVertices>> povsForGraph = new HashMap<Graph, Set<PairOfVertices>>();
  
  /**
   * @return the singleton instance of the traveller
   */
  public static GraphTraveller getTraveller()
  {
    if (INSTANCE == null) {
      INSTANCE = new GraphTraveller();
    }
    return INSTANCE;
  }
  
  /**
   * private constructor
   */
  private GraphTraveller() {}

  /**
   * @param threeNodesGraph
   * @return
   */
  public Set<PairOfVertices> findAllPairsOfVerticesOn(Graph graph)
  {
    if (graph == null)
    {
      throw new GraphTravellerException("Input graph was null");
    }

    Set<PairOfVertices> allPovs = null;

    if (povsForGraph.containsKey(graph))
    {
      allPovs = povsForGraph.get(graph);
    }

    else
    {

      allPovs = new LinkedHashSet<PairOfVertices>();

      Set<Vertex> vertices = findAllVerticesOn(graph);
      for (Vertex v : vertices)
      {
        for (Vertex v2 : vertices)
        {
          if (!v.equals(v2))
          {
            PairOfVertices pov = new PairOfVertices(v, v2);
            allPovs.add(pov);
          }
        }
      }
    }

    /* store it in cache */
    povsForGraph.put(graph, allPovs);

    return allPovs;
  }
  
  public Set<Vertex> findAllVerticesOn(Graph graph){
    
    Set<Vertex> vertices = new HashSet<Vertex>();
    
    VertexIterator vi = graph.vertices();
    
    while (vi.hasNext()) {
      Vertex v = vi.nextVertex();
      log.finest("Vertex : "+v);
      vertices.add(v);
    }
    return vertices;
  }

  /**
   * @param graph
   * @param vertexId
   * @return
   */
  public Vertex locateVertexById(Graph graph, String vertexId)
      throws VertexNotFoundException
  {

    Vertex located = null;
    
    //for each vertex
    final VertexIterator vertexIterator = graph.vertices();
    while (vertexIterator.hasNext())
    {
      Vertex v = vertexIterator.nextVertex();
      final String vertexDecoration = (String) v.element();
      if (vertexDecoration.equalsIgnoreCase(vertexId))
      {
        located = v;
        break; //vertex found, stop searching
      }
    }
    if (located == null)
    {
      throw new VertexNotFoundException("Cannot locate vertex with id: "
          + vertexId + " on the input graph");
    }
    return located;
  }

  /**
   * @param graph the input graph
   * @param sourceVertexId the source vertex id
   * @param destVertexId the destination vertex id
   * @return the shortest path between source and dest
   * @throws VertexNotFoundException if source or destination vertex are not
   *             found on the input graph
   */
  public Path computeShortestPathBetween(Graph graph, String sourceVertexId,
      String destVertexId) throws VertexNotFoundException
  {

    Vertex source = locateVertexById(graph, sourceVertexId);
    Vertex dest = locateVertexById(graph, destVertexId);
    return computeShortestPathBetween(graph, source, dest);

  }

  /**
   * @param graph the input graph
   * @param sourceVertex the source vertex
   * @param destVertex the destination vertex
   * @return the shortest path between source and destination
   */
  public Path computeShortestPathBetween(Graph graph, Vertex sourceVertex,
      Vertex destVertex)
  {
    ShortestPathExecutor executor = new ShortestPathExecutor();
    return executor.executeAndReturn(graph, sourceVertex, destVertex);
  }

  /**
   * For every pair of vertices in the graph, compute the shortest path, returning a map filled with the computed informations.
   * 
   * @param graph the input graph
   * @return 
   */
  public Map<PairOfVertices, Path> computeShortestPathBetweenAnyPairOfVerticesIn(
      Graph graph)
  {

    Map<PairOfVertices, Path> shortestPathForPairs = new HashMap<PairOfVertices, Path>();

    final Set<PairOfVertices> allPairs = findAllPairsOfVerticesOn(graph);

    for (PairOfVertices pov : allPairs)
    {

      Path sp = computeShortestPathBetween(graph, pov.v1, pov.v2);

      shortestPathForPairs.put(pov, sp);

    }

    return shortestPathForPairs;
  }

}
